import h5py
import os
import shutil
import pathlib
import numpy as np
import xml.etree.ElementTree as et
import datetime
import pyedflib

def format_comment(comment: str) -> str:
    lines = comment.strip().split("\n")
    formatted = ", ".join(f"{lines[i]}:{lines[i+1]}" for i in range(0, len(lines), 2) if i+1 < len(lines))
    return formatted

def parse_xml(xml_str):
    root = et.fromstring(xml_str)  # Parse XML
    parsed_data = {}
    for child in root:
        parsed_data[child.tag] = child.text.strip() if child.text else None
    return parsed_data

class EEG():
    def __init__(self, hdf_filepath=None):
        self.filepath = None
        self.sampling_frequency = None
        self.channel_num = None
        self.channel_data = None
        self.feature_onsets = None
        self.feature_descriptions = None
        self.session_run = None
        self.session_comment = None
        self.acquisition_device = None
        self.acquisition_unit = None
        
        if hdf_filepath != None:
            self.read_hdf5(hdf_filepath)
    
        
    
    def print_features(self):
        assert(len(self.onsets) == len(self.descs)) # 
        for idx, onset in enumerate(self.onsets):
            print(f"Feature {idx} @ {self.onsets[idx]}:{self.descs[idx]}")
        
    def print(self):
        print("EEG : ", self.filepath)
        print("sampling_frequency : ", self.sampling_frequency)  
        print("channel_num : ", self.channel_num )
        print("channel_data : ", self.channel_data.shape)
        print("feature_onsets : ", self.feature_onsets)
        print("feature_descriptions : ", self.feature_descriptions) 
        print("session_run :", self.session_run)
        print("session_comment :", self.session_comment)
        print("acquisition_device : ", self.acquisition_device)
        print("acquisition_unit : ", self.acquisition_unit)
        
    def read_hdf5(self, filepath):
        print("Reading HDF5 file : ", filepath)
        
        self.filepath = filepath
        self.hdf = h5py.File(filepath, mode="r+")
        print(self.hdf.keys)
        # Handle 'RawData' containing the acutal channeld ata
        rawdata_group = self.hdf["RawData"]
        self.channel_data = np.array(rawdata_group["Samples"]).transpose()
        print("Samples : ", self.channel_data.shape)
        
        # 'AsynchronData' : This contains Features / Markers
        async_data_group = self.hdf ["AsynchronData"] 
        if async_data_group.get("Time"):
            self.feature_onsets = np.array(async_data_group["Time"]).T[0]
            self.feature_descriptions = np.array(async_data_group["TypeID"]).T[0]
            print("feature_onsets : ", self.feature_onsets)
            print("feature_descriptions : ", self.feature_descriptions) 
        else:
            self.feature_onsets = []
            self.feature_descriptions = []
        # 'AcquisitionTaskDescription' 
        acquisition_task_desc = parse_xml(rawdata_group["AcquisitionTaskDescription"].asstr()[0])
        channel_properties = acquisition_task_desc.get("ChannelProperties")
        self.channel_num = int(acquisition_task_desc.get("NumberOfAcquiredChannels"))
        self.sampling_frequency = float(acquisition_task_desc.get("SamplingFrequency"))
        print("channel_properties : ", channel_properties)
        print("channel_num : ", self.channel_num )
        print("sampling_frequency : ", self.sampling_frequency)  
        
        #   'DAQDeviceDescription
        daq_desc = parse_xml(rawdata_group["DAQDeviceDescription"].asstr()[0])
        self.acquisition_unit = daq_desc.get("Unit") # micro Volts
        self.acquisition_device = daq_desc.get("Name")
        print("acquisition_unit : ", self.acquisition_unit)
        print("acquisition_device : ", self.acquisition_device)
        
        #   'SessionDescription'
        session_desc = parse_xml(rawdata_group["SessionDescription"].asstr()[0])
        print(session_desc)
        self.session_run = session_desc.get("Run")
        self.session_comment = session_desc.get("Comment")
        print("session_run :", self.session_run)
        print("session_comment :", self.session_comment)
        
        #   'SubjectDescription'
        subject_desc = parse_xml(rawdata_group["SubjectDescription"].asstr()[0])
        #comment = format_comment(subject_desc.get("Comment"))
        #first_name = subject_desc.get("FirstName")
        #last_name = subject_desc.get("LastName")
        #print("comment : ", comment)
        #print("name : ", first_name, " ", last_name)
        
        self.hdf.close()
        # 'SavedFeatures' # NOTE : This is not important
        #saved_features_group = self.hdf["SavedFeatues"]
        #saved_features_keys = saved_features_group.keys()
        #features_num = saved_features_group["NumberOfFeatures"].astype('int32')[0]
        #print("features_num : ", features_num)

        # 'Version' # NOTE : This is not important
        #version_group = self.hdf["Version"]
        #version_keys = version_group.keys()
        #version = version_group["Version"].asstr()[0]
        #print("version : ", version)

        #   'DAQDeviceCapabilities' # NOTE : No important information here as far as I can tell
        #daw_capabilities = parse_xml(rawdata_group["DAQDeviceCapabilities"].asstr()[0])
        
        return self
    
    # TODO : Instead of manual new filepath append something to the already existing filepath
    def write_hdf5(self, new_filepath:str, append_to_filepath:str) -> "EEG":
        
        old_filepath = self.filepath #  What file to copy
        
        filename, suffix = old_filepath.split(".") 
        temp_filepath = filename + "_temporary." + suffix # Where to create a copy
        
        shutil.copy(old_filepath, temp_filepath) # Copy file
        print(f"Write HDF5 : {old_filepath} copied, new filepath : {new_filepath}")

        # Loads the temporary file, alters it
        hdf = h5py.File(temp_filepath, mode="r+")
        print("Write HDF5 : Loaded {hdf} in read+ mode.")

        #Replace features
        async_group = hdf["AsynchronData"]
        del async_group["Time"]
        del async_group["TypeID"] 
        async_group.create_dataset("Time", data=np.array(self.feature_onsets).reshape(-1, 1))
        async_group.create_dataset("TypeID", data=np.array(self.feature_descriptions).reshape(-1, 1))
        
        # Overwrite channel data
        rawdata_group = hdf["RawData"]
        del rawdata_group["Samples"]
        rawdata_group.create_dataset("Samples", data=self.channel_data.T, dtype="f4")
        
        hdf.close()
       
        # Then relocates and renames the file
        if os.path.exists(new_filepath):
            ans = input(f"{new_filepath} already exists. Do you want to overwrite it? [Y / N] : ")
            if ans.capitalize() == "Y":
                shutil.move(temp_filepath, new_filepath)
                print(f"Wrote HDF5 to {new_filepath}")
            else:
                print("Write canceled....")
                os.remove(temp_filepath)
                return
        else:
            os.rename(temp_filepath, new_filepath) # Change file name & path
            print(f"Wrote HDF5 to {new_filepath}")

        return EEG(new_filepath)
    
    # TODO : Implment
    def write_edf(self,):
        
        
        pass
    
    def read_edf(self,):
        
        pass
    
    def trim(self, cut_from_start:float=0, cut_from_end:float=0):
        """Trims the EEG channel data """
        
        sample_num = self.channel_data.shape[1]
        start = int(cut_from_start * self.sampling_frequency)
        end = sample_num - int(cut_from_end * self.sampling_frequency)

        if start < 0 or end <= len(self.feature_onsets):
            print("Trim parameters out of bounds, trimming was NOT conducted...")
            return 
        
        self.channel_data = self.channel_data[:, start:end]

        valid_indices = [i for i, onset in enumerate(self.feature_onsets) if start <= onset < end]
        self.feature_onsets = np.array([self.feature_onsets[i] - start for i in valid_indices])
        self.feature_descriptions = [self.feature_descriptions[i] for i in valid_indices]
    
    def trim_from_features(self, cut_from_first_feature:float=5, cut_from_last_feature:float=10) -> None:
        """Trims the EEG channel data before the first feature and after the last"""
        print(f"Trimming EEG : [ First feature - {cut_from_first_feature}, Last feature + {cut_from_last_feature} ]")

        first_feature_onset = self.feature_onsets[0]
        last_feature_onset = self.feature_onsets[len(self.feature_onsets)-1]

        sample_num = self.channel_data.shape[1]
        start = int(first_feature_onset - (cut_from_first_feature * self.sampling_frequency))
        end = int(last_feature_onset + (cut_from_last_feature * self.sampling_frequency))

        if start < 0 or end <= len(self.feature_onsets):
            print("Trim parameters out of bounds, trimming was NOT conducted...")
            return 
        
        self.channel_data = self.channel_data[:, start:end]
        
        valid_indices = [i for i, onset in enumerate(self.feature_onsets) if start <= onset < end]
        self.feature_onsets = np.array([self.feature_onsets[i] - start for i in valid_indices])
        self.feature_descriptions = [self.feature_descriptions[i] for i in valid_indices]
    
    def preprocess(self, filtering=True, normalization=True, ):
        
        pass
    


from neuropipeline.fnirs import fNIRS
from neuropipeline.eeg import EEG

eeg = EEG()
snirf = fNIRS()
snirf.read_snirf("C:/dev/neuro-glial-analysis/data/Subject01/Trial 3 - Supination/2025-03-24_003.snirf")
snirf.feature_epochs(2, 2, 5)
exit()
snirf.wl_to_od()
snirf.od_to_hb()
snirf.write_snirf("C:/dev/test_snirf.snirf")'

import numpy as np
import matplotlib.pyplot as plt
from preprocessing.filter import butter_bandpass_filter, notch_filter
from analysis.frequencies import compute_psd, compute_fft
from sklearn.decomposition import PCA, FastICA
from scipy.integrate import simpson
from wrappers.eeg import EEG

def remove_movement_artifacts(data, n_components=5):
    pca = PCA(n_components=n_components)
    transformed = pca.fit_transform(data.T)
    transformed[:, :n_components] = 0  # Remove first few components
    return pca.inverse_transform(transformed).T

def trim(eeg:EEG, cut_from_first_feature=0, cut_from_last_feature=0 ):
    print(f"Trimming EEG : [ First feature - {cut_from_first_feature}, Last feature + {cut_from_last_feature} ]")
    s_freq = eeg.sampling_frequency
    
    first_feature_onset = eeg.feature_onsets[0]
    last_feature_onset = eeg.feature_onsets[len(eeg.feature_onsets)-1]
    
    sample_num = eeg.channel_data.shape[1]
    
    start = int(first_feature_onset - (cut_from_first_feature * s_freq))
    end = int(last_feature_onset + (cut_from_last_feature * s_freq))
    
    print(start)
    print(end)
    eeg.channel_data = eeg.channel_data[:, start:end]
    
    print("orignianl onsets: ", eeg.feature_onsets)
    
    valid_indices = [i for i, onset in enumerate(eeg.feature_onsets) if start <= onset <= end]
    eeg.feature_onsets = np.array([eeg.feature_onsets[i] - start for i in valid_indices])
    eeg.feature_descriptions = [eeg.feature_descriptions[i] for i in valid_indices]
    
    print("trimmed onsets: ", eeg.feature_onsets)
    return eeg



def bandpass_channels(data, s_freq, lowcut, highcut, order):
    filtered = np.zeros((data.shape))
    
    for idx in range(len(data)):
        filtered[idx] = butter_bandpass_filter(data[idx], lowcut, highcut, s_freq, order)
    
    return filtered
def common_average_reference_filter(data):
    average_reference = np.mean(data, axis=0)
    averaged = data - average_reference
    return averaged

def normalize(data, znorm=False, baseline_correction=False, log=False):
    normalized = data
    # Z-Normalization
    if znorm:
        normalized = (data - np.mean(data, axis=1, keepdims=True)) / np.std(data, axis=1, keepdims=True)
    
    # Baseline Correction
    if baseline_correction:
        baseline = data[:, :2560]  # Assume first 50 samples are pre-stimulus
        baseline_mean = np.mean(baseline, axis=1, keepdims=True)
        normalized = data - baseline_mean
    # Log Transformation
    if log:
        normalized = np.log1p(np.abs(data))
        
    return normalized

def motion_correction(data):
    motion_corrected = data
    n = 5
    #pca = PCA(n_components=n)
    #ica = FastICA(n_components=n)
    #transformed_pca = pca.fit_transform(filtered.transpose())  # Transpose so PCA works on timepoints
    #transformed_ica = ica.fit_transform(filtered.transpose())  # Transpose so PCA works on timepoints
    #transformed_pca[:, :n] = 0  
    #transformed_ica[:, :2] = 0  
    #cleaned_pca = pca.inverse_transform(transformed_pca).transpose()  # Transpose back
    #cleaned_ica = ica.inverse_transform(transformed_ica).transpose()  # Transpose back
    return motion_corrected
    

def preprocess(eeg:EEG, bandpass=True, normalization=True, ica=True, rerefernce=True): 
    
    # Copy data to modify
    processed_data = eeg.channel_data.copy()
    
    #
        
    if bandpass: # Apply Bandpass filtering -> 1 to 100 Hz, 50, 60, 100 Hz notch filter
        for idx in range(len(processed_data)):
            filtered_time_series = butter_bandpass_filter(processed_data[idx], 8, 30, eeg.sampling_frequency, 5)
            notched = notch_filter(filtered_time_series, eeg.sampling_frequency, freqs=[50, 60, 100])
            processed_data[idx] = notched
              
    
    if ica:
        ica = FastICA(n_components=processed_data.shape[0])
        transformed_ica = ica.fit_transform(processed_data)#.transpose())  # Transpose so PCA works on timepoints
        transformed_ica[:, :12] = 0  
        cleaned_ica = ica.inverse_transform(transformed_ica)#.transpose()  # Transpose back
        eeg.preprocessing_history.append({'ica': {}})
        processed_data = cleaned_ica
        
        
    # Common Average Refrence 
    if rerefernce:
        processed_data = common_average_reference_filter(processed_data)
        
    # Motion Correction -> Not implemented
    #processed_data = motion_correction(processed_data) 
    #eeg.preprocessing_history.append({'motion_correction': {}})
    
    # Normalization
    if normalization:
        processed_data = normalize(processed_data, znorm=True)
    


    # Replace channel data
    eeg.channel_data =  processed_data
    return eeg

band_ranges_spec = {
        "Delta (0.5-4 Hz)": (0.5, 4),
        "Theta (4-8 Hz)": (4, 8),
        "Alpha (8-12 Hz)": (8, 12),
        "Beta (12-30 Hz)": (12, 30),
        "Gamma (30-100 Hz)": (30, 100),
    }

band_power_colors = ['blue', 'green', 'orange', 'red', 'purple']

def compute_band_power(spectra, freqs):
    """
    Computes band power for predefined EEG frequency bands.
    
    Args:
        spectra (numpy.ndarray): Power spectral density values.
        freqs (numpy.ndarray): Corresponding frequency values.
    
    Returns:
        dict: Band powers for Delta, Theta, Alpha, Beta, and Gamma bands.
    """
    spectra_res = spectra[1] - spectra[0] 
    spectral_density = simpson(freqs, dx=spectra_res)
    
    band_powers = {}
    for band, (low, high) in band_ranges_spec.items():
        
        band_mask = np.logical_and(spectra >= low, spectra <= high) # TRUE when the sepctra Hz is within low and high,  FALSE where spectra is below low or above high
        band_powers[band]  = simpson(freqs[band_mask], dx=spectra_res) / spectral_density
       
    return band_powers   
'


import numpy as np
import matplotlib.pyplot as plt
from enum import Enum
from mne.io import read_raw_snirf
from mne_nirs.io import write_raw_snirf
from snirf import validateSnirf
from mne.preprocessing.nirs import optical_density, beer_lambert_law,  temporal_derivative_distribution_repair
from mne import Annotations
from scipy.signal import butter, sosfreqz, sosfiltfilt, iirnotch, filtfilt, welch, detrend
from scipy.stats import pearsonr



def TDDR(signal, sample_rate):
    # This function is the reference implementation for the TDDR algorithm for
    #   motion correction of fNIRS data, as described in:
    #
    #   Fishburn F.A., Ludlum R.S., Vaidya C.J., & Medvedev A.V. (2019).
    #   Temporal Derivative Distribution Repair (TDDR): A motion correction
    #   method for fNIRS. NeuroImage, 184, 171-179.
    #   https://doi.org/10.1016/j.neuroimage.2018.09.025
    #
    # Usage:
    #   signals_corrected = TDDR( signals , sample_rate );
    #
    # Inputs:
    #   signals: A [sample x channel] matrix of uncorrected optical density data
    #   sample_rate: A scalar reflecting the rate of acquisition in Hz
    #
    # Outputs:
    #   signals_corrected: A [sample x channel] matrix of corrected optical density data
    signal = np.array(signal)
    if len(signal.shape) != 1:
        for ch in range(signal.shape[1]):
            signal[:, ch] = TDDR(signal[:, ch], sample_rate)
        return signal

    # Preprocess: Separate high and low frequencies
    filter_cutoff = .5
    filter_order = 3
    Fc = filter_cutoff * 2/sample_rate
    signal_mean = np.mean(signal)
    signal -= signal_mean
    if Fc < 1:
        fb, fa = butter(filter_order, Fc)
        signal_low = filtfilt(fb, fa, signal, padlen=0)
    else:
        signal_low = signal

    signal_high = signal - signal_low

    # Initialize
    tune = 4.685
    D = np.sqrt(np.finfo(signal.dtype).eps)
    mu = np.inf
    iter = 0

    # Step 1. Compute temporal derivative of the signal
    deriv = np.diff(signal_low)

    # Step 2. Initialize observation weights
    w = np.ones(deriv.shape)

    # Step 3. Iterative estimation of robust weights
    while iter < 50:

        iter = iter + 1
        mu0 = mu

        # Step 3a. Estimate weighted mean
        mu = np.sum(w * deriv) / np.sum(w)

        # Step 3b. Calculate absolute residuals of estimate
        dev = np.abs(deriv - mu)

        # Step 3c. Robust estimate of standard deviation of the residuals
        sigma = 1.4826 * np.median(dev)

        # Step 3d. Scale deviations by standard deviation and tuning parameter
        r = dev / (sigma * tune)

        # Step 3e. Calculate new weights according to Tukey's biweight function
        w = ((1 - r**2) * (r < 1)) ** 2

        # Step 3f. Terminate if new estimate is within machine-precision of old estimate
        if abs(mu - mu0) < D * max(abs(mu), abs(mu0)):
            break

    # Step 4. Apply robust weights to centered derivative
    new_deriv = w * (deriv - mu)

    # Step 5. Integrate corrected derivative
    signal_low_corrected = np.cumsum(np.insert(new_deriv, 0, 0.0))

    # Postprocess: Center the corrected signal
    signal_low_corrected = signal_low_corrected - np.mean(signal_low_corrected)

    # Postprocess: Merge back with uncorrected high frequency component
    signal_corrected = signal_low_corrected + signal_high + signal_mean

    return signal_corrected

def butter_bandpass(lowcut, highcut, fs, freqs=512, order=3):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    sos = butter(order, [low, high], btype='band', output='sos')
    w, h = sosfreqz(sos, worN=None, whole=True, fs=fs)
    return sos, w, h

def butter_bandpass_filter(time_series, lowcut, highcut, fs, order):
    sos, w, h = butter_bandpass(lowcut, highcut, fs, order=order)
    y = sosfiltfilt(sos, time_series)
    return np.array(y)

def notch_filter(data, sfreq, freqs=[50, 60]):
    """Apply notch filters at specified frequencies."""
    for freq in freqs:
        b, a = iirnotch(freq, 30, sfreq)
        data = filtfilt(b, a, data, axis=-1)
    return data

from scipy.signal import freqz, sosfreqz

def compute_fft(time_series, fs, freq_limit:float|None):
    # Compute FFT
    N = len(time_series)  # Length of the signal
    fft_result = np.fft.fft(time_series)
    fft_freq = np.fft.fftfreq(N, d=1/fs)#/fs)  # Frequency axis

    # Take the positive half of the spectrum
    positive_freqs = fft_freq[:N // 2]
    positive_spectrum = np.abs(fft_result[:N // 2]) * (2 / N)  # Normalize for one-sided
    
    if freq_limit is None:
        return positive_freqs, positive_spectrum

    # Filter frequencies to only include up to freq_limit
    indices = positive_freqs <= freq_limit
    limited_freqs = positive_freqs[indices]
    limited_spectrum = positive_spectrum[indices]
    return limited_freqs, limited_spectrum

def compute_psd(time_series, fs, freq_limit:float|None):
    # Compute FFT
    freqs, spectrum = compute_fft(time_series, fs, freq_limit)
    # Normalize to get power spectral density
    psd = np.square(spectrum) / (fs * len(time_series))  
    # Double the PSD for one-sided spectrum (except at DC and Nyquist)
    psd[1:] = 2 * psd[1:]
    return freqs, psd

class fnirs_data_type(Enum):
    Wavelength = "Wavelength"
    OpticalDensity = "Optical Density"
    HemoglobinConcentration = "Hemoglobin Concentration"

WL = fnirs_data_type.Wavelength
OD = fnirs_data_type.OpticalDensity
CC = fnirs_data_type.HemoglobinConcentration

class fNIRS():
    def __init__(self, filepath=None): 
        self.type = WL
        self.snirf = None 
        
        self.sampling_frequency = None
        self.channel_names = None
        self.channel_data = None
        self.channel_num = None
        
        self.feature_onsets = None
        self.feature_descriptions = None
        
        if filepath != None:
            self.read_snirf(filepath)
    
    def print(self):
        print("sampling_frequency : ", self.sampling_frequency, " Hz")
        print("channel_num : ", self.channel_num)
        print("channel_data : ", self.channel_data.shape)
        print("channel_names : ", self.channel_names)
        print("feature_onsets : ", self.feature_onsets)
        print("feature_descriptions : ", self.feature_descriptions)
        
    def read_snirf(self, filepath):
        print(f"Reading SNIRF from {filepath}")
        result = validateSnirf(filepath)
        print("valid : ", result.is_valid())
        self.snirf = read_raw_snirf(filepath)
        snirf = read_raw_snirf(filepath)
        # fNIRS info
        info = self.snirf.info
        self.sampling_frequency = float(info["sfreq"])
        self.channel_names = info["ch_names"]
        self.channel_data = np.array(self.snirf.get_data())
        self.channel_num = int(info["nchan"])
        # Features
        annotations = self.snirf._annotations
        self.feature_onsets = np.array(annotations.onset, dtype=float)
        self.feature_descriptions = np.array(annotations.description, dtype=int)
        self.feature_durations = np.array(annotations.duration, dtype=float)

        self.channel_dict = None
        
    def update_snirf_object(self):
        
        # Overwrite channel data
        self.snirf._data = self.channel_data
        
        # Fix Annotations
        new_annotations = Annotations(onset=self.feature_onsets,
                                      duration=self.feature_durations,
                                      description=self.feature_descriptions)
        self.snirf.set_annotations(new_annotations)
        
    def get_channel_dict(self):
        self.channel_dict = {}
        for i, channel_name in enumerate(self.channel_names):
            
            source_detector = channel_name.split()[0]
            wavelength = channel_name.split()[1]
            
            if source_detector not in self.channel_dict:
                self.channel_dict[source_detector] = {"HbO" : None, 
                                                 "HbR" : None
                                                 }
            
            channel_data = self.channel_data[i] 
            
            if wavelength == "HbR".lower() or wavelength == "760":
                self.channel_dict[source_detector]["HbR"] = channel_data
                
            if wavelength == "HbO".lower() or wavelength == "850":
                self.channel_dict[source_detector]["HbO"] = channel_data
        return self.channel_dict
    
    def split(self):
        """
        Splits the channel data into HbO and HbR
        """
        assert(len(self.channel_data) == len(self.channel_names))
        
        hbo_channels = []
        hbo_names = []
        hbr_channels = []
        hbr_names = []
        for i, channel_name in enumerate(self.channel_names): 
            parts = channel_name.split()
            
            assert(len(parts) == 2)
            
            source_detector, wavelength = parts[0], parts[1].lower()
        
            if wavelength == "hbr" or wavelength == "760":
                hbr_channels.append(self.channel_data[i] )
                hbr_names.append(source_detector)
                
            if wavelength == "hbo" or wavelength == "850":
                hbo_channels.append(self.channel_data[i] )
                hbo_names.append(source_detector)
        
        # Into numpy arrays!
        hbo_channels = np.array(hbo_channels)
        hbr_channels = np.array(hbr_channels)
        
        return hbo_channels, hbo_names, hbr_channels, hbr_names
    
    def write_snirf(self, filepath):
        write_raw_snirf(self.snirf, filepath)
        print(f"Wrote SNIRF to {filepath}")
        result = validateSnirf(filepath)
        print("valid : ", result.is_valid())
        

    def to_optical_density(self, use_inital_value=False):
        """
        Converts raw light intensity data to optical density. \nif use_inital_value is False then this function mimicks MNE's optical_density function.

        Parameters:
            raw_data (numpy.ndarray): 2D array [channels x samples] of raw light intensities.

        Returns:
            optical_density (numpy.ndarray): Converted optical density data.
        """
        
        # We need to compare how different this result is from the mne ones is
        
        if self.type != WL:
            print(f"sNIRF type is {self.type}, cannot convert to {OD}!")
            return
        
        if use_inital_value: # Use I_0 according to Dans 2019
            measured_intensity = self.channel_data
            initial_intensity = measured_intensity[:, 0]

            # Avoid division by zero
            safe_intensity = np.clip(measured_intensity, a_min=1e-12, a_max=None)
            safe_initial = np.clip(initial_intensity[:, np.newaxis], a_min=1e-12, a_max=None)

            od = -np.log(safe_intensity / safe_initial)
        else: # Use mean according to MNE.nirs
            
            data = np.abs(self.channel_data)  # Take absolute to avoid negative intensities

            # Replace zeros by the smallest positive value per channel
            min_nonzero = np.min(np.where(data > 0, data, np.inf), axis=1, keepdims=True)
            data = np.maximum(data, min_nonzero)

            # Normalize each channel by its mean
            means = np.mean(data, axis=1, keepdims=True)
            normalized = data / means

            # Apply natural log and invert sign
            od = -np.log(normalized)
        
        self.channel_data = od
        self.type = OD
        ch_dict = {}
        for ch_name in self.channel_names:
            ch_dict[ch_name] = "fnirs_od"
            
        self.snirf.set_channel_types(mapping=ch_dict)

    def to_hemoglobin_concentration(self):
        if self.type != OD:
            print(f"sNIRF type is {self.type}, cannot convert to {CC}!")
            return 
        
        self.update_snirf_object()
        hb = beer_lambert_law(self.snirf)
        
        self.snirf = hb
        
        # TODO : Do we actually need to reread all this?
        info = self.snirf.info
        self.channel_names = info["ch_names"]
        self.channel_data = np.array(self.snirf.get_data())
    

    def feature_epochs(self, feature_description, tmin, tmax):
        
        onsets = [] # Fill with the onsets
        print(self.feature_descriptions)
        print(self.feature_onsets)
        for i, desc in enumerate(self.feature_descriptions):
            if desc == feature_description:
                onsets.append(self.feature_onsets[i])
        print("feature : ", feature_description, f" ({len(onsets)})")
        print("onsets : ", onsets)
        
        exit()
        for i, channel_name in enumerate(self.channel_dict):
            
            pass
        
        
        pass
    
    
    def remove_features(self, features_to_remove):
        """Removes features by description match."""
    
        indices_to_keep = [
            i for i, desc in enumerate(self.feature_descriptions) 
            if desc not in features_to_remove
        ]

        self.feature_onsets = np.array([self.feature_onsets[i] for i in indices_to_keep])
        self.feature_descriptions = np.array([self.feature_descriptions[i] for i in indices_to_keep])
        self.feature_durations = np.array([self.feature_durations[i] for i in indices_to_keep])

        print(f"Removed {len(self.feature_descriptions) - len(indices_to_keep)} features. Remaining: {len(self.feature_descriptions)}.")
        self.update_snirf_object()
        
    def trim_from_features(self, cut_from_first_feature:float=5, cut_from_last_feature:float=10) -> None:
        
        first_seconds = self.feature_onsets[0]
        last_seconds = self.feature_onsets[-1]
        
        first_frame = first_seconds * self.sampling_frequency
        last_frame = last_seconds * self.sampling_frequency
        
        start_seconds = first_seconds - cut_from_first_feature
        end_seconds = last_seconds + cut_from_last_feature
        
        start_frames = int(first_frame - (cut_from_first_feature * self.sampling_frequency))
        end_frames = int(last_frame + (cut_from_last_feature * self.sampling_frequency))
        
        print(f"Trimming From Features {self.channel_data.shape} : [ {start_seconds} : {end_seconds} ] / [ {start_frames} : {end_frames} ]")
        assert(start_frames >= 0)
        if end_frames < self.channel_data.shape[1]:
            end_frames = self.channel_data.shape[1]-1
            
        self.channel_data = self.channel_data[:, start_frames:end_frames]
        
        valid_indices = [i for i, onset in enumerate(self.feature_onsets) if start_seconds <= onset < end_seconds]
        self.feature_onsets = np.array([self.feature_onsets[i] - start_seconds for i in valid_indices])
        self.feature_descriptions = np.array([self.feature_descriptions[i] for i in valid_indices])
        self.feature_durations = np.array([self.feature_durations[i] for i in valid_indices])
        # OVERWRITE THE .SNIRF
        self.update_snirf_object()

    def bandpass_channels(self, low_freq=0.01, high_freq=0.1, order=5):
        """
        Applies a digital bandpass filter to all channels. Returns filtered snirf object. 

        Args:
            snirf (RawSNIRF) : RawSNIRF object
            l_freq : Lowcut frequency, the lower edge of passband
            h_freq : Highcut frequency, the high edge of passband  
            n : Filter order, higher means small transition band

        Returns:
            filtered (RawSNIRF) : New RawSNIRF object with filtered channels
        """
        for i, channel in enumerate(self.channel_data):  # Iterate over each channel
            self.channel_data[i] = butter_bandpass_filter(channel, low_freq, high_freq, self.sampling_frequency, order)
        
# Plan to replace this
# class Preprocessor : def set_bandpass_settings(self, l_freq, h_freq, n):
# preprocessor __init__(, optical_density=True, hemoglobin_concentration=True, motion_correction=True, temporal_filtering=True, detrending=True, normalization=True )
# and so on        
# class Preprocessor:
#     def __init__(self, optical_density=True, hemoglobin_concentration=True, 
#                  motion_correction=True, 
#                  temporal_filtering=True, detrending=True, normalization=True)  :
        
        
#         self.optical_density = optical_density
#         self.hemoglobin_concentration = hemoglobin_concentration
#         self.motion_correction = motion_correction
#         self.temporal_filtering = temporal_filtering
#         self.detrending = detrending
#         self.normalization = normalization
        
#     def get_settings(self):
#         return {
#             "optical_density" : self.optical_density,
#             "hemoglobin_concentration" : self.hemoglobin_concentration,
#             "motion_correction" : self.motion_correction,
#             "temporal_filtering" : self.temporal_filtering,
#             "detrending" : self.detrending,
#             "normalization" : self.normalization
#         }
        
#     def preprocess(self, fnirs:fNIRS):
#         fnirs.preprocess(
#             optical_density=self.optical_density,
#             hemoglobin_concentration=self.hemoglobin_concentration,
#             motion_correction=self.motion_correction,
#             temporal_filtering=self.temporal_filtering,
#             detrending=self.detrending,
#             normalization=self.normalization
#         )
        
              
    def preprocess(self, optical_density=True, hemoglobin_concentration=True, motion_correction=True, temporal_filtering=True, detrending=True, normalization=True):
        if optical_density:
            self.to_optical_density(use_inital_value=False)

        if hemoglobin_concentration:
            self.to_hemoglobin_concentration()

        if motion_correction:
            # Apply TDDR motion correction
            for i, channel in enumerate(self.channel_data):
                self.channel_data[i] = TDDR(channel, self.sampling_frequency)
            self.snirf._data = self.channel_data
            
        if temporal_filtering:
            lowcut = 0.01
            highcut = 0.1
            order = 15
            self.bandpass_channels(lowcut, highcut, order)
            
        if detrending:
            # Apply linear detrending to remove slow drifts
            for i, channel in enumerate(self.channel_data):
                self.channel_data[i] = detrend(channel, type='linear')
            self.snirf._data = self.channel_data
            
        if normalization:
            mean_vals = np.mean(self.channel_data, axis=1, keepdims=True)  # Compute mean per channel
            std_vals = np.std(self.channel_data, axis=1, keepdims=True)  # Compute standard deviation per channel

            std_vals[std_vals == 0] = 1

            normalized_channels = (self.channel_data - mean_vals) / std_vals  # Apply Z-Normalization

            self.channel_data = normalized_channels
            self.snirf._data = normalized_channels
            
    def plot_channels(self,):
        
        hbo_data, hbo_names, hbr_data, hbr_names = self.split()
    
        plt.figure(figsize=(12, 8))

        # Plot HbO time series
        plt.subplot(2, 2, 1)
        for i, ch in enumerate(hbo_data):
            plt.plot(ch, label=f"{hbo_names[i]}")
        plt.title("HbO Time Series")
        plt.legend()

        # Plot HbR time series
        plt.subplot(2, 2, 2)
        for i, ch in enumerate(hbr_data):
            plt.plot(ch, label=f"{hbr_names[i]}")
        plt.title("HbR Time Series")
        plt.legend()

        # Plot HbO Power Spectral Density
        plt.subplot(2, 2, 3)
        for i, ch in enumerate(hbo_data):
            freqs, spectra = compute_psd(ch, self.sampling_frequency, int(self.sampling_frequency/2))
            plt.plot(freqs, spectra)
        plt.title("HbO : Power Spectral Density")
        plt.xlabel("Frequency [Hz]")
        plt.ylabel("PSD [V²/Hz]")
        plt.legend()

        # Plot HbR Power Spectral Density
        plt.subplot(2, 2, 4)
        for i, ch in enumerate(hbr_data):
            freqs, spectra = compute_psd(ch, self.sampling_frequency, int(self.sampling_frequency/2))
            plt.plot(freqs, spectra)
        plt.title("HbR : Power Spectral Density")
        plt.xlabel("Frequency [Hz]")
        plt.ylabel("PSD [V²/Hz]")
        plt.legend()

        plt.tight_layout()
        plt.show()